/*

 compat.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "compat.h"
#include "driver.h"
#include "ideserv.h"
#include "path.h"
#include <dirent.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>

#define OPEN_RONLY 0
#define OPEN_STATUS 2
#define OPEN_WONLY 4
#define OPEN_ERR 6

int openfile_compat(const struct io_ops_s *io_ops, struct buffer_s *buff, FILE *lf, const struct arguments_s *arguments) {
    c64c_t cmd[256];
    unsigned char status=OPEN_ERR;
    unsigned char channel,fp;
    unsigned int length=0;
    struct buffer_s *buffer;
    int r;
    const c64c_t *outname;

    r=io_ops->getb();if (r < 0) return 1;
    channel=r & 0x0f;
    buffer=&buff[channel];

    fp=0;
    while (fp<255) {
        r=io_ops->getb();
        if (r < 0) {
            if (arguments->mode == M_ETHERNET) break;
            return 1;
        }
        cmd[fp]=r;
        if (!cmd[fp]) break;
        fp++;
    }
    cmd[fp]=0;

    alarm(0);

    {
        if (buffer->file) {
            fclose(buffer->file);
            buffer->file=NULL;
        }
        buffer->mode=CM_CLOSED;

        {
            c64c_t name[17], type[4]={'*',0};
            unsigned char mode;
            char lname[1000],outpath[1020];
            struct dirent *ep;
            DIR *dp;
            int found=0;

            resolv_path(cmd, outpath, &outname, NULL, lf, arguments->relaxfiletypes);
            convertfilename(outname, ',', name, type, &mode);

            dp=opendir(outpath[0]?outpath:".");
            if (!dp) {
                errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);
                if (io_ops->sendb(2)) return 1;
                if (arguments->mode == M_ETHERNET && io_ops->sendb(2)) return 1;
                if (io_ops->flush()) return 1;
                return 0;
            }
            if (type[0]==0) strcpy((char *)type,(channel==1)?"PRG":"*");

            initdirentry(0);
            while ((ep=readdir(dp)))
            {
                DIRECTORY_ENTRY_T dirent;
                char lname2[2020];

                strncpy(lname,ep->d_name,999);
                strcpy(lname2,outpath);
                if (outpath[0]) strncat(lname2,"/", 999);
                strncat(lname2,ep->d_name,999);

                if (createdirentry(ep->d_name, lname2, &dirent, arguments->relaxfiletypes)) continue;

                if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_CLOSED | A_NORMAL)) continue;

                length = dirent.filesize[0] + dirent.filesize[1] * 256 + dirent.filesize[2] * 65536 + dirent.filesize[3] * 65536 * 256;

                dirent.filesize[0]=0;
                if (!matchname(dirent.name, name)) continue;
                dirent.time[0]=0;
                if (!matchname(dirent.filetype,type)) continue;

                found=1;
                break;
            }
            closedir(dp);
            initdirentry(1);
            status=OPEN_ERR;

            if (!found) {
                mbstate_t ps;
                int f=0,f2=0,g=0;
                memset(&ps, 0, sizeof(ps));
                while (name[f] && f < 16) {
                    f2 += c64toascii(lname + f2, name[f], &ps);
                    f++;
                }
		if (!arguments->relaxfiletypes || memcmp(type,"PRG",3)) {
            	    lname[f2++]=',';
            	    while (type[g] && g < 3) {
                        f2 += c64toascii(lname + f2, type[g], &ps);
                        g++;
            	    }
		}
                lname[f2]=0;
            }

            if (outpath[0]) strcat(outpath,"/");
            strcat(outpath,lname);

            fprintf(lf,"Open [%d]: \"%s\": ",channel,outpath);

            if (!name[0]) {
                seterror(34,0);goto vege;
            }

            {
                if (!found) {
                    seterror(62,0);
                } else {
#ifndef WIN32
                    buffer->fd=open(outpath, O_RDONLY, 0);
#else
                    buffer->fd=open(outpath, O_RDONLY|O_BINARY, 0);
#endif
                    if (buffer->fd<0) errtochannel15(1); else {
                        status=OPEN_RONLY;//ok
                        buffer->file=fdopen(buffer->fd, "rb");
                        buffer->mode=CM_COMPAT;
                        buffer->filesize=length;
                    }
                }
            }
        }
    }
vege:
    if (io_ops->sendb(status)) return 1;
    if (arguments->mode == M_ETHERNET && io_ops->sendb(status)) return 1;
    if (io_ops->flush()) return 1;
    return 0;
}

int readfile_compat(const struct io_ops_s *io_ops, struct buffer_s *buff, FILE *lf, const struct arguments_s *arguments) {
    unsigned char channel,status=0;
    unsigned short int bytes,i;
    struct timeval start, end;
    unsigned long long l=0,l2;
    struct buffer_s *buffer;
    int r;

    r=io_ops->getb();if (r < 0) return 1;
    channel=r & 0x0f;
    buffer=&buff[channel];
    r=io_ops->getb();if (r < 0) return 1;
    bytes=r;
    if (arguments->mode == M_ETHERNET) {
        if (!bytes) bytes=256;
    } else {
        r=io_ops->getb();if (r < 0) return 1;
        bytes|=r << 8;
        if (io_ops->getb() < 0) return 1;
    }
    fprintf(lf,"Read [%d] %d bytes\n",channel,bytes);
    if (arguments->mode != M_ETHERNET) {
        if (io_ops->sendb(0)) return 1;
    }
    if (buffer->mode!=CM_COMPAT) {
        if (io_ops->sendb(0)) return 1;
        if (io_ops->sendb(0)) return 1;
        if (arguments->mode != M_ETHERNET && io_ops->sendb(0)) return 1;
        if (io_ops->flush()) return 1;
        return 0;
    } else {
        unsigned long itt;

        if (!buffer->data) {
            buffer->data=(unsigned char *)malloc(65536);
            if (!buffer->data) return 1;//out of memory?!
        }

        l=bytes=fread(buffer->data,1,bytes?bytes:65536,buffer->file);
        if (feof(buffer->file)) status=0x42; else
        {
            itt=ftell(buffer->file);
            fgetc(buffer->file);
            if (feof(buffer->file)) status=0x40;
            fseek(buffer->file,itt,SEEK_SET);
        }
    }
    if (arguments->mode == M_ETHERNET) {
        if (io_ops->sendb(bytes>0)) return 1;
        if (io_ops->sendb(bytes)) return 1;
        for (i=0;i<bytes;i++) io_ops->sendb(buffer->data[bytes-i-1]);
        if (bytes & 1) if (io_ops->sendb(status)) return 1;
        if (io_ops->sendb(status)) return 1;
        if (io_ops->sendb(status)) return 1;
        if (io_ops->flush()) return 1;
    } else {
        if (io_ops->sendb(~bytes)) return 1;
        if (io_ops->sendb(~bytes >> 8)) return 1;
        if (io_ops->flush()) return 1;

        gettimeofday(&start,NULL);
        if (bytes) {
            if (io_ops->sendbytes(bytes, buffer->data)) return 1;
        } else {
            if (io_ops->sendbytes(32768, buffer->data)) return 1;
            if (io_ops->sendbytes(32768, buffer->data+32768)) return 1;
        }
        if (io_ops->sendb(status)) return 1;//fprintf(lf,"Status: %02X\n",status);
        if (io_ops->flush()) return 1;
        gettimeofday(&end,NULL);
        l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
        if (l2) {
            l=l*1000000/l2;
            fprintf(lf,"Read %d bytes at %lld bytes/seconds\n",bytes,l);
        }
    }
    return 0;
}

int closefile_compat(const struct io_ops_s *io_ops, struct buffer_s *buff, FILE *lf, const struct arguments_s *arguments) {
    unsigned char channel;
    struct buffer_s *buffer;
    int r;

    r=io_ops->getb();if (r < 0) return 1;
    channel=r & 0x0f;
    buffer=&buff[channel];
    if (arguments->mode == M_ETHERNET) {
        if (io_ops->sendb(0)) return 1;
        if (io_ops->sendb(0)) return 1;
    } else {
        if (io_ops->getb() < 0) return 1;
        if (io_ops->sendb(0)) return 1;
    }
    if (io_ops->flush()) return 1;
    fprintf(lf,"Close [%d]\n",channel);
    if (buffer->mode==CM_COMPAT) {
        if (buffer->file) fclose(buffer->file);
        buffer->mode=CM_CLOSED;buffer->file=NULL;
        if (buffer->data) {
            free(buffer->data);
            buffer->data=NULL;
        }
    }
    return 0;
}
